/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * mdisk.c-- will be used as a memory-based block device
 * or "memory-disk"
 *
 * create one "memory disk", which may be partitioned in
 * the same way an ordinary disk is
 *
 * (from param.h), we note that the "sector" size is DEV_BSIZE
 * this is generally 512 bytes.  Any mention of "block" in ths following
 * code is relative to 512-byte sectors.
 *
 * Kevin Fall
 * November 10, 1992	(started and finished)
 * kfall@ucsd.edu
 */

#include "../h/param.h"
#include "../h/types.h"
#include "../h/devio.h"
#include "../h/ioctl.h"
#include "../h/errno.h"
#include "../fs/ufs/fs.h"
#include "../h/buf.h"
#include "../h/kmalloc.h"

#ifdef notdef
#define	MDISK_NSECTS	40960	/* # sectors in this disk */
#endif
#define	MDISK_NSECTS	1024	/* 1/2 a MB */
#define	MDISK_SIZE	(MDISK_NSECTS*DEV_BSIZE)

static struct devget	dev_info = {
	DEV_DISK,	/* category */
	DEV_NB,		/* bus type == none */
	DEV_RAMDISK,	/* memory-based disk (interface) */
	DEV_RAMDISK,	/* memory-based disk (device) */
	0,		/* adapt_num */
	0,		/* nexus_num */
	0,		/* bus_num */
	0,		/* ctlr_num */
	0,		/* remote ctlr_num */
	0,		/* slave_num */
	"mdisk",	/* dev_name (8 bytes max) */
	0,		/* unit num */
	0,		/* driver soft error count */
	0,		/* driver hard error count */
	DEV_DONE,	/* status mask (generic), done  */
	0L
};

struct bogus {
  	unsigned long   dev_size; /* # blocks in the user area (#pc) */
        unsigned short  ntracks; /* # tracks/cyl (#nt) */
        unsigned short  nsectors; /* # sects/trk (#ns) */
        unsigned short  ncylinders; /* # cyls (#nc) */
        unsigned long   attributes; /* attributes */
};

static struct bogus mdisk_geom = {
		(u_long) MDISK_NSECTS,	/* # blks in user area */
		(u_short) 1,		/* tracks/cyl */
		(u_short) (MAXBSIZE / DEV_BSIZE),        /* sects/trk (16) */
		(u_short) (MDISK_SIZE / MAXBSIZE),	/* # cyls */
		(u_long) 0			/* attributes */
};

static struct pt mdisk_pt = {
	PT_MAGIC,	/* magic # */
	PT_VALID,	/* info is valid info */
	{
		0,			/* a partition-empty */
		(daddr_t) 0,		/* offset */
	},
	{
		0,			/* a partition-empty */
		(daddr_t) 0,		/* offset */
	},
	{
		MDISK_SIZE / DEV_BSIZE,	/* c partition--full */
		(daddr_t) 0,		/* offset */
	}
	/* .. rest is filled in as 0 */
};

static caddr_t	mdisk_created = NULL;	/* holds address of "disk" */
caddr_t	create_mdisk();
void free_mdisk();

/*
 * open routine for bdevsw
 */

mdisk_bopen(dev, flag)
	dev_t	dev;
	int	flag;
{
#ifdef MDISK_DEBUG
uprintf("mdisk_bopen(%d,%d), mdisk_created = 0x%x\n",
dev, flag, mdisk_created);
#endif

	if (mdisk_created == NULL) {
		if ((mdisk_created = create_mdisk(MDISK_SIZE)) ==
		    NULL) {
#ifdef MDISK_DEBUG
uprintf("mdisk_bopen: fs create failed\n");
#endif
			return (ENOMEM);	/* sort of */
		}

#ifdef MDISK_DEBUG
uprintf("mdisk_bopen: fs created at 0x%x\n", mdisk_created);
#endif
	}
	return (0);	/* success */
}

mdisk_bclose(dev, flag)
	dev_t	dev;
	int	flag;
{
	if (mdisk_created == NULL)
		return (0);	/* shouldn't happen */

	free_mdisk(mdisk_created);
	mdisk_created = NULL;
	return (0);
}

mdisk_cclose(dev, flag)
	dev_t dev;
	int flag;
{
	return (0);
}

/*
 * open routine for cdevsw
 */

mdisk_copen(dev, flag)
	dev_t	dev;
	int	flag;
{
	return (mdisk_bopen(dev, flag));
}


/*
 * ioctl routine used in bdevsw and cdevsw
 */

mdisk_ioctl(dev, cmd, data, flag)
	dev_t	dev;
	int	cmd;
	caddr_t	data;
	int	flag;
{

#ifdef MDISK_DEBUG
uprintf("mdisk_ioctl(dev=%d,cmd=%d,data=0x%x,flag=%d)\n",
dev, cmd, data, flag);
#endif

	switch (cmd) {

	case DEVIOCGET:	/* used for ufs mount */
	{
		struct devget *dget = (struct devget *) data;
		bzero((caddr_t) dget, sizeof(struct devget));
		*dget = dev_info;	/* struct copy */
		dget->category_stat = DEV_DISKPART;
		break;
	}

	case DEVGETGEOM:
	{
		DEVGEOMST *devg = (DEVGEOMST *) data;
		*devg = * ((DEVGEOMST *) &mdisk_geom);	/* struct copy */
		break;
	}

	case DIOCGETPT:
	case DIOCDGTPT:
	{
		/* get partition or default partition table */
		struct pt *pt = (struct pt *) data;
		*pt = mdisk_pt;	/* struct copy */
		break;
	}

	case DIOCSETPT:
	{
		struct pt *pt = (struct pt *) data;
		int err;

		if (!suser())
			return (EACCES);

		if ((err = ptcmp(dev, &mdisk_pt, pt)) != 0)
			return (err);

		mdisk_pt = *pt;	/* copy in new pt from user */

		ssblk(dev, pt);	/* sync in-core superblock */
		mdisk_pt.pt_valid = PT_VALID;
		break;
	}

	default:
		return (ENXIO);

	}	/* switch */

#ifdef MDISK_DEBUG
uprintf("mdisk_ioctl returning ok\n");
#endif

	return (0);
}

/*
 * return partition size
 */

mdisk_bsize(dev)
	dev_t	dev;
{
	int partition;

	partition = minor(dev) & 0x07;
	return (mdisk_pt.pt_part[partition].pi_nblocks);
}

extern int biodone();

mdisk_strategy(bp)
	register struct buf *bp;
{
	int	partition = minor(bp->b_dev) & 0x07;
	int	maxblk;
	long	blkno = bp->b_blkno;
	int	nsectors = (bp->b_bcount + (DEV_BSIZE - 1)) >> 9;

	u_char	(*p)[DEV_BSIZE] = (u_char (*) [DEV_BSIZE])
		mdisk_created;	/* sector pointer */

#ifdef MDISK_DEBUG
buf_debug(bp);
#endif

fstrace_rec("REQ", bp);

	maxblk = mdisk_pt.pt_part[partition].pi_nblocks;

	/* check for disk existence */

	if (mdisk_created == NULL) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENODEV;	/* out of space */
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		return;
	}

	/* check for size ok */

#ifdef MDISK_DEBUG
uprintf("mdisk_strategy: part=%d, blkno=%ld, maxblk = %d, b_count=%d\n",
partition, blkno, maxblk, bp->b_bcount);
#endif

	if ((blkno < 0) || (blkno+nsectors > maxblk)) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENOSPC;	/* out of space */
		bp->b_resid = bp->b_bcount;
		biodone(bp);
		return;
	}

	p += blkno;	/* p points to data of interest */


	if (bp->b_flags & B_READ) {
		bcopy((u_char *) p, bp->b_un.b_addr, bp->b_bcount);
		biodone(bp);
		return;
	}

	if ((bp->b_flags & B_READ) == 0) {
		bcopy(bp->b_un.b_addr, (u_char *) p, bp->b_bcount);
		biodone(bp);
		return;
	}
	return;
}

static caddr_t
create_mdisk(sz)
	int sz;
{
	caddr_t		fsptr;

	static u_char bigbuf[MDISK_SIZE];

	/* KM_ALLOC(fsptr, caddr_t, sz, KM_TEMP, KM_CLEAR); */

	return ((caddr_t) bigbuf);
}

static void
free_mdisk(ptr)
	caddr_t ptr;
{
	/* KM_FREE(ptr, KM_TEMP); */
	return;
}

/*
 * note: physio routine will take the user's uio structure
 * and point the b_addr field in the buf header into it.
 * hence, when the strategy routine is called, the io is
 * performed directy to the user's buffer
 */

static struct buf	tmpbuf;

mdisk_write(dev, uio)
	dev_t dev;
	struct uio *uio;
{
#ifdef MDISK_DEBUG
uprintf("mdisk_write: calling physio\n");
#endif
	return (physio(mdisk_strategy, &tmpbuf, dev, B_WRITE, minphys, uio));
}

mdisk_read(dev, uio)
	dev_t dev;
	struct uio *uio;
{
#ifdef MDISK_DEBUG
uprintf("mdisk_read: calling physio\n");
#endif
	return (physio(mdisk_strategy, &tmpbuf, dev, B_READ, minphys, uio));
}

