
/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/types.h"
#include "../h/mbuf.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/gnode_common.h"
#include "../fs/ufs/ufs_inode.h"
#include "../h/gnode.h"
#include "../fs/ufs/fs.h"
#include "../h/conf.h"
#include "../h/buf.h"
#include "../h/mount.h"
#include "../h/kernel.h"
#include "../h/fs_types.h"
#include "../h/file.h"
#include "../h/kmalloc.h"
#include "../h/exit_actn.h"
#include "../h/exec.h"
#include "../h/proc.h"
#include "../h/workstation.h"
#include "../h/inputdevice.h"
#include "../h/wsdevice.h"

#include "../net/net/route.h"
#include "../net/net/if.h"

#include "../net/netinet/in.h"
#include "../net/netinet/in_pcb.h"
#include "../net/netinet/in_systm.h"
#include "../net/netinet/in_var.h"

#include "splice.h"

#define	SPLICE_DEBUG	1


/*
 * splice begin:
 *	check args for appropriateness
 *	allocate/setup splice descriptor
 *	set up handlers for process and splice read/write
 *	kick things off by calling splice_rw
 */

int
splice_begin(fin, fout)
	struct file *fin, *fout;
{
	register struct gnode *src;
	register struct gnode *dest;

	register struct socket *srcsock;
	register struct socket *destsock;

	struct splice_descriptor *sp;
	off_t srcoffset, destoffset;
	struct buf *bp;
	int rval, needclean;

	srcoffset = fin->f_offset;
	destoffset = fout->f_offset;

	/* check type (i.e. socket versus inode */
	if (!splice_types_supported(fin->f_type, fout->f_type)) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: type unsupp, failure 01");
#endif
		u.u_error = EOPNOTSUPP;
		return 0;
	}

	/*
	 * GET GNODE or SOCKET POINTERS FROM OPEN FILE TABLE
	 */

		/* Src Side */

	if (ISINODE(fin)) {
		/* Do special stuff related to gnodes */
		if ((src = (struct gnode *) fin->f_data) == NULL) {
			u.u_error = EBADF;
			return 0;
		}
		if (!splice_filetype_supported(src)) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: filetypetype unsupp, failure 02");
#endif
			u.u_error = EOPNOTSUPP;
			return 0;
		}
		if (src->g_size > u.u_rlimit[RLIMIT_FSIZE].rlim_cur) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: exceeded user settable filesize limit\n");
#endif
			if (u.u_procp->p_progenv == A_BSD)
				psignal(u.u_procp, SIGXFSZ);
			u.u_error = EFBIG;
			return 0;
		}
		/* if there's no data, we don't do any work */
		if (ISREG(src)) {
			if (src->g_size == 0)
				return 0;
			else if (srcoffset != 0) {
				u.u_error = EOPNOTSUPP;
#ifdef SPLICE_DEBUG
cprintf("splice_begin: not supported nonzero offsets");
#endif
				return 0;
			}
		}

	} else if (ISSOCK(fin)) {
		if ((srcsock = (struct socket *) fin->f_data) == NULL) {
			u.u_error = EBADF;
			return 0;
		}
		if (!splice_sockettype_supported(srcsock)) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: socktype unsupp, failure 03");
#endif
			u.u_error = EOPNOTSUPP;
			return 0;
		}
	} else {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: unknown unsupp, failure 04");
#endif
		u.u_error = EOPNOTSUPP;
		return 0;
	}

		/* Dest side */

	if (ISINODE(fout)) {
		/* Do special stuff related to gnodes */
		if ((dest = (struct gnode *)fout->f_data) == 0) {
			u.u_error = EBADF;
			return 0;
		}
		if (!splice_filetype_supported(dest)) {
			u.u_error = EOPNOTSUPP;
			return 0;
		}

		if (ISREG(dest) && destoffset != 0) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: not supported nonzero offsets");
#endif
			u.u_error = EOPNOTSUPP;
			return 0;
		}
	} else if (ISSOCK(fout)) {
		if ((destsock = (struct socket *)fout->f_data) == NULL) {
			u.u_error = EBADF;
			return 0;
		}
		if (!splice_sockettype_supported(destsock)) {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: destsock unsupp, failure 05");
#endif
			u.u_error = EOPNOTSUPP;
			return 0;
		}
	} else {
#ifdef SPLICE_DEBUG
cprintf("splice_begin: unk unsupp, failure 06");
#endif
		u.u_error = EOPNOTSUPP;
		return 0;
	}

	/*
	 * GENERIC FOR ALL SPLICES...
	 */

	sp = splice_alloc();

	if (sp == NULL) {
		u.u_error = ENOMEM;
		return 0;
	}

	sp->sp_state |= SPLICE_ACTIVE;

	/*
	 * set up initial values for splice descriptor
	 */

	sp->sp_srcfile = fin;

	if (ISINODE(fin)) {
		sp->sp_srcgp = src;
		sp->sp_srcoffset = 0;
	} else if (ISSOCK(fin)) {
		sp->sp_srcsock = srcsock;
		sp->sp_srcoffset = 0;
	} else
		panic("unknown src type");

	sp->sp_destfile = fout;

	if (ISINODE(fout)) {
		sp->sp_destgp = dest;
	} else if (ISSOCK(fout)) {
		sp->sp_destsock = destsock;
	} else
		panic("unknown src type");

	sp->sp_owrites = 0;
	sp->sp_oreads = 0;

	/*
	 * *must* assign proc ptr before calling
	 * splice_establish_procexit()
	 */

	sp->sp_proc = u.u_procp;

#ifdef notdef
this code goes in when switching to fully async version!
if (splice_establish_procexit(sp) < 0) {
	u.u_error = ENOMEM;
	return 0;
}
#endif

	/*
	 * adjust reference counts/establish locks
	 */

#ifdef notdef
	smp_lock(&fin->f_lk, LK_RETRY);
	smp_lock(&fout->f_lk, LK_RETRY);
#endif

	if (ISINODE(fin) && ISREG(src)) {
		sp->sp_srcfile->f_count++;
		gfs_lock(src);
	}

	if (ISINODE(fout) && ISREG(dest)) {
		sp->sp_destfile->f_count++;
		gfs_lock(dest);
	}

	/*
	 * DOIT
	 */

	if (ISINODE(fin) && ISINODE(fout)) {

		/*
		 * Inode -> Inode
		 */

		if (!(ISREG(src) && ISREG(dest))) {
			u.u_error = EOPNOTSUPP;
			return 0;
		}

		/*
		 * first, figure out all the src and dest phys block #s
		 */

		if (splice_map(sp) < 0) {
			return (-1);
		}

		/*
		 * next, loop moving the data
		 */

		rval = splice_rw(sp);

	} else if (ISSOCK(fin) && ISINODE(fout)) {
		/*
		 * SOCKET -> Inode (reg or char special file)
		 */
		struct inpcb *inp = (struct inpcb *) srcsock->so_pcb;
		dev_t	dev;

		inp->inp_splicep = (caddr_t) sp;
		/* for now, support network to char special device */
		if (ISCHR(dest)) {
			/* dev is the maj/min #s specified by the inode */
			dev = dest->g_rdev;	/* only for c-special files */
			if (csplicesw[major(dev)].putfn)
				sp->sp_cdevput = csplicesw[major(dev)].putfn;
			else {
				u.u_error = EOPNOTSUPP;
				return 0;
			}
		} else {
			u.u_error = EOPNOTSUPP;
			return 0;
		}

	} else if (ISINODE(fin) && ISSOCK(fout)) {
		/*
		 * Inode -> SOCKET
		 */


		if (ISCHR(src)) {
			/* SUPPORT FOR CHAR-SPECIAL FILES */

			dev_t dev;

			/* dev is the maj/min #s specified by the inode */
			dev = src->g_rdev;	/* only for c-special files */

			if ((major(dev) > SPLICE_NCDEVS)  ||
			   (csplicesw[major(dev)].getfn == NULL)) {
			
				u.u_error = EOPNOTSUPP;
				return 0;
			}
			sp->sp_cdevget = csplicesw[major(dev)].getfn;
		} else {
			/* SUPPORT FOR REGULAR FILES GOES HERE */
			u.u_error = EOPNOTSUPP;
			return 0;
		}
		/* for now, kick off by calling get fn */
		(*sp->sp_cdevget)(sp, NULL, 0);	/* no data ptr */
		/* u.u_error = EOPNOTSUPP; */
		return 0;
	} else if (ISSOCK(fin) && ISSOCK(fout)) {
		/*
		 * socket -> socket (new code, 10/94)
		 *
		 * splice_net_upcall will be invoked when an
		 * appropriate packet arrives, so not much to do here
		 */

                struct inpcb *inp = (struct inpcb *) srcsock->so_pcb;

		/* both sockets get a copy of the splice descriptor */
                inp->inp_splicep = (caddr_t) sp;
		((struct inpcb *)destsock->so_pcb)->inp_splicep = (caddr_t) sp;
		sp->sp_state |= SPLICE_NO_OUTCKSUM;

		/* FALLTHRU, will wait below */

	}


	/*
	 * Be careful: if the following sleep code is
	 * removed, it would be wrong to call splice_finish
	 * right away, because there may still be some pending
	 * junk.  If splice_finish is called too early, splice_read
	 * or splice_write will crash with a vaddr page fault or
	 * tlb miss when it tries to refer to the blockmap[lbn],
	 * as the heap memory has been freed already
	 */

	while ((sp->sp_state&SPLICE_ACTIVE)) {
		sp->sp_state |= SPLICE_WANTED;
		sleep((caddr_t) sp, PZERO+1);
	}
	splice_finish(sp);
	return 0;
}


/*
 * splice_finish()
 *
 *	1) set "update access time" on src gnode
 *	2) update fields in dest gnode
 *	3) unlock dest and src gnodes
 *	4) remove spinlock on file pointers
 *
 *	this routine is called by splice_rw or in the
 *	process exit handler
 *	when an EOF condition has occured on the source
 */

splice_finish(sp)
	struct splice_descriptor *sp;
{
	struct socket *dest_sock;
	struct socket *src_sock;

	struct gnode *dest_gp;
	struct gnode *src_gp;

	if (sp->sp_srcfile->f_type == DTYPE_INODE) {
		src_gp = sp->sp_srcgp;
		src_gp->g_flag |= GACC;
		gfs_unlock(src_gp);
	}

	if (sp->sp_destfile->f_type == DTYPE_INODE) {
		dest_gp = sp->sp_destgp;
		dest_gp->g_size = src_gp->g_size;
		dest_gp->g_flag |= GUPD|GCHG;
		dest_gp->g_mode &= ~(GSUID|GSGID); /* no setuid/gid */
		gfs_unlock(dest_gp);
	}


#ifdef SPLICE_DEBUG
cprintf("splice_finish: unlocking file pointers\n");
#endif

	sp->sp_srcfile->f_count--;
	sp->sp_destfile->f_count--;

#ifdef notdef
	smp_unlock(&sp->sp_destfile->f_lk); 
	smp_unlock(&sp->sp_srcfile->f_lk); 
#endif

	splice_free(sp);

	return (0);

}

/*
 * splice_alloc()
 *	allocate the memory for a splice_descrptor and
 *	put it on the systemwide list of active splices
 */

struct splice_descriptor *sys_splices;	/* linked list, all splices */
struct splice_descriptor *
splice_alloc()
{
	struct splice_descriptor *sp;
	int s;

	KM_ALLOC(sp, struct splice_descriptor *,
		sizeof(*sp), KM_TEMP, KM_CLEAR);

	if (sp == NULL) {
#ifdef SPLICE_DEBUG
cprintf("splice_alloc: KM_ALLOC failed for splice descriptor\n");
#endif
		return (NULL);
	}

	/* insert into front of doubly-linked list */
	s = splbio();
	sp->sp_next = sys_splices;
	if (sp->sp_next)
		sp->sp_next->sp_prev = sp;
	sp->sp_prev = NULL;
	sys_splices = sp;
	splx(s);

	return (sp);
}

/*
 * splice_free()
 *
 *	1) pull the splice descriptor from the system splice list
 *	2) free the src and destination blockmap arrays
 *	3) free the splice descriptor
 *
 * 	called by splice_finish()
 */

int
splice_free(sp)
	struct splice_descriptor *sp;
{
	int s = splbio();

	if (!sp) {
		splx(s);
		panic("splice_free: null sd");
	}

	if (sp->sp_prev)
		sp->sp_prev->sp_next = sp->sp_next;
	if (sp->sp_next)
		sp->sp_next->sp_prev = sp->sp_prev;
	if (sp == sys_splices)
		sys_splices = sp->sp_next;
	splx(s);

	if (sp->sp_srcfile->f_type == DTYPE_INODE && sp->sp_srcblockmap)
		KM_FREE(sp->sp_srcblockmap, KM_TEMP);

	if (sp->sp_destfile->f_type == DTYPE_INODE &&
	    sp->sp_destblockmap)
		KM_FREE(sp->sp_destblockmap, KM_TEMP);

	KM_FREE(sp, KM_TEMP);
	return (0);
}

/*
 * find the first splice descriptor for a given process ptr
 */

struct splice_descriptor *
find_splice_proc(p)
	struct proc *p;
{
	struct splice_descriptor *sp = sys_splices;
	int s = splbio();

	while (sp) {
		if (p == sp->sp_proc) {
			splx(s);
			return (sp);
		}
		sp = sp->sp_next;
	}
	splx(s);
	return (NULL);
}

/*
 * hook up the process exit handler to the given process
 */

int
splice_establish_procexit(sp)
	struct splice_descriptor *sp;
{
	struct proc *p = sp->sp_proc;
	struct exit_actn *xp = NULL;
	int s;

#ifdef SPLICE_DEBUG
cprintf("establish_procexit: sp=0x%x, procp=%x\n",sp,p);
#endif

	KM_ALLOC(xp, struct exit_actn *,
		sizeof(*xp), KM_EXIT_ACTN, KM_NOARG);

	if (xp == NULL) {
#ifdef SPLICE_DEBUG
cprintf("splice_establish_procexit: out of memory\n");
#endif
		return (-1);
	}

	/* insert at beginning of process exit handlers list */ 
	xp->xa_func = splice_exit_handler;
	s = splhigh();
	xp->xa_next = u.u_exitp;
	u.u_exitp = xp;
	splx(s);

	sp->sp_xa = xp;	/* for find_splice_xa() */

	return (0);
}

/*
 * this function is invoked by exit() when a process terminates
 * if it returns non-zero, the exit_actn descriptor is *not*
 * freed
 *
 * the idea is to find *all* splices a process may have outstanding
 * and wait for them to terminate.  Only then should the process
 * be permitted to exit
 */

int
splice_exit_handler(xa)
	struct exit_actn *xa;
{
	struct splice_descriptor *sp;

#ifdef SPLICE_DEBUG
cprintf("exit_handler: xa=0x%x, sp=%x\n",
xa, find_splice_xa(xa)); 
#endif

loop:
	if ((sp = find_splice_xa(xa)) == NULL)
		return (0);

	while (sp->sp_owrites) {
		sp->sp_state |= SPLICE_WANTED;
#ifdef SPLICE_DEBUG
cprintf("exit_handler: sleeping on 0x%x\n", sp);
#endif
		sleep((caddr_t) sp, PZERO+1);
	}
		
	splice_finish(sp);
	goto loop;
}

/*
 * find the first splice descriptor given an exit action
 *	this function is needed because during process exit,
 *	the exit action handlers are only passed a pointer
 *	to themselves, and we must find a way of finding
 *	all the splices a process might have
 */

struct splice_descriptor *
find_splice_xa(xa)
	struct exit_actn *xa;
{
	struct splice_descriptor *sp = sys_splices;

	int s = splbio();

	while (sp) {
		if (xa == sp->sp_xa) {
			splx(s);
			return (sp);
		}
		sp = sp->sp_next;
	}
	splx(s);
	return (NULL);
}

/*
 * splice_map:
 *	given a gnode and whether we're reading or
 *	writing it, figure out all of the phys blk #s
 *	for the file.
 *
 *	Note: calls to bmap may block, so only call this
 *	routing when there's a process context around.
 */

int
splice_map(sp)
	register struct splice_descriptor *sp;
{
	register struct gnode *srcgp = sp->sp_srcgp;
	register struct gnode *destgp = sp->sp_destgp;

	struct fs *srcfs = FS(srcgp);
	struct fs *destfs = FS(destgp);

	dev_t srcdev = (dev_t) srcgp->g_dev;
	dev_t destdev = (dev_t) destgp->g_dev;

	long srcbsize = srcfs->fs_bsize;
	long destbsize = destfs->fs_bsize;

	register daddr_t lbn, pbn, *blockmap;

	daddr_t src_lastlbn, dest_lastlbn;
	int offset, n, sync_flag = 0, *sptr = &sync_flag;

	/* get last logical blk # for files (may vary by fs) */
	/* make sure we the size of the src gnode for both! */
	src_lastlbn = lblkno(srcfs, srcgp->g_size);
	dest_lastlbn = lblkno(destfs, srcgp->g_size);

	sp->sp_destlastlbn = dest_lastlbn;

	if (srcbsize != destbsize) {
#ifdef SPLICE_DEBUG
cprintf("splice_map: filesystems have differing block size\n");
#endif
		u.u_error = EOPNOTSUPP;
		return (-1);
	}

	KM_ALLOC(blockmap, daddr_t *,
		(int)((src_lastlbn + 1)*sizeof(daddr_t)),
		KM_TEMP, KM_CLEAR);

	if (blockmap == (daddr_t *)NULL) {
#ifdef SPLICE_DEBUG
cprintf("splice_map: couldn't allocate space for srcblockmap\n");
#endif
		u.u_error = EFBIG; 
		return (-1);
	}

	sp->sp_srcblockmap = blockmap;

	KM_ALLOC(blockmap, daddr_t *,
		(int)((dest_lastlbn + 1)*sizeof(daddr_t)),
		KM_TEMP, KM_CLEAR);


	if (blockmap == (daddr_t *)NULL) {
#ifdef SPLICE_DEBUG
cprintf("splice_map: couldn't allocate space for destblockmap\n");
#endif
		u.u_error = EFBIG;
		return (-1);
	}

	sp->sp_destblockmap = blockmap;

#ifdef SPLICE_DEBUG
cprintf("splice_map(src):lastlbn=%d\n", src_lastlbn);
#endif

	/*
	 * map the source file
	 */

	blockmap = sp->sp_srcblockmap;
	lbn = 0;
	do {
		pbn = ufs_bmap(srcgp, (int)lbn, B_READ, 0, (int *) 0);
		if ((int)pbn <= 0)
			panic("splice_map: bmap src");
		*blockmap++ = pbn;
	} while (lbn++ < src_lastlbn);

#ifdef SPLICE_DEBUG
cprintf("splice_map(src-physblknos):");
{
	register i; 
	for (i = 0;i <= (int)src_lastlbn; i++)
		cprintf(" %d ", sp->sp_srcblockmap[i]);
	cprintf("\n");
}
#endif


#ifdef SPLICE_DEBUG
cprintf("splice_map(dest):lastlbn=%d\n", dest_lastlbn);
#endif

	/*
	 * map the destination file
	 * use a special version of ufs_bmap which doesn't cause
	 * a bunch of writes with the (B_FORCEWRITE|B_ASYNC) bits
	 * turned on, as the ordinary ufs_bmap routine does
	 */

	blockmap = sp->sp_destblockmap;
	offset = 0;
	do {
		lbn = lblkno(destfs, offset);
		n = MIN((unsigned) destbsize, srcgp->g_size - offset);
		if ((offset +n) > destgp->g_size) {
			destgp->g_size = offset + n;
		}
		pbn = splice_ufs_bmap(destgp, (int) lbn,
			B_WRITE, n, (int *) 0);
		if ((int) pbn <= 0)
			panic("splice_map: bmap destination");
		*blockmap++ = pbn;
		offset += n;
	} while (lbn < dest_lastlbn);

#ifdef SPLICE_DEBUG
cprintf("splice_map(dest-physblknos):");
{
	register i; 
	for (i = 0;i <= (int)dest_lastlbn; i++)
		cprintf(" %d ", sp->sp_destblockmap[i]);
	cprintf("\n");
}
#endif

	return (0);
}

/*
 * splice_bmap is basically bmap, but modified in such
 * a way that the data blocks allocated when a file is
 * grown are not written to disk.  All other data
 * (meta-data, dir blocks, etc) remain as they have always
 * been
 */

splice_ufs_bmap(gp,bn,rwflg,size,sync)
register struct gnode *gp;
daddr_t bn;	/* virtual block of file to un-map */
int rwflg;	/* purpose of call */
int size;	/* supplied only when rwflg == B_WRITE */
int *sync;
MOST OF THIS FUNCTION WAS DERIVED FROM THE ULTRIX VERSION
SEE YOUR OWN VERSION FOR REFERENCE
{
#ifdef SPLICE
			else {
				brelse(bp);
			}
#else
			else
				bdwrite(bp);
#endif
....
#ifdef SPLICE
			else {
				brelse(bp);
			}
#else
			else
				bdwrite(bp);
#endif
...
#ifdef SPLICE
			else {
				brelse(nbp);
			}
#else
			else
				bdwrite(nbp);
#endif
}

/*
 * return 0 (false) if not supported
 * return 1 (true) if supported
 */

int
splice_types_supported(intype, outtype)
	int intype;
	int outtype;
{
	/* accept file-to-file copies */
	if ((intype == DTYPE_INODE) && (outtype == DTYPE_INODE))
		return 1;

	/* accept network-to-file copies */  
	if ((intype == DTYPE_SOCKET) && (outtype == DTYPE_INODE))
		return 1;

	/* accept file-to-network copies */  
	if ((intype == DTYPE_INODE) && (outtype == DTYPE_SOCKET))
		return 1;

	/* accept network-to-network copies */
	if ((intype == DTYPE_SOCKET) && (outtype == DTYPE_SOCKET))
		return 1;
	

	/* don't accept whatever else we've been given */

#ifdef SPLICE_DEBUG
cprintf("splice_types_supported: no supp. for itype=%d, otype=%d\n",
intype, outtype);
#endif
	return 0;
}

/*
 * splice_filetype_supported
 *	for gnodes, determine if we may splice this thing...
 *	For now, don't accept special devices or nonlocal files
 */

int
splice_filetype_supported(fl)
	struct gnode *fl;
{
	/* support character-dev files */
	if (ISCHR(fl)) {
		return (1);
	}

	if (!ISREG(fl)) {

#ifdef SPLICE_DEBUG
cprintf("splice_begin: non-regular files not supported\n");
#endif
		return 0;
	}

	if (!ISLOCAL(fl->g_mp)) {

#ifdef SPLICE_DEBUG
cprintf("splice_begin: files must be locally mounted\n");
#endif
		return 0;
	}

	return 1;
}

/*
 * insist any sockets are of AF_INET, bound, and DGRAM
 */

int
splice_sockettype_supported(sock)
	struct socket *sock;
{
	register struct inpcb *inp;
	
	/* are we in internet domain? */
	if (sock->so_proto->pr_domain->dom_family != PF_INET)
		return 0;

	/* are we a datagram */
	if (sock->so_type != SOCK_DGRAM)
		return 0;

	/* are we bound? */
	inp = (struct inpcb *) sock->so_pcb;
	if (!inp->inp_lport)
		return 0;

	return 1;
}

int wsget(), wsput();

/*
 * table of functions for getting and putting info
 * from char devs involved in a splice.  Indexed by
 * major device number
 */

struct splice_cdev csplicesw[SPLICE_NCDEVS] = {
	(int (*)())	NULL,		/* 0 */
	(int (*)())	NULL,		/* 0 */
	(int (*)())	NULL,		/* 1 */
	(int (*)())	NULL,		/* 1 */
	(int (*)())	NULL,		/* 2 */
	(int (*)())	NULL,		/* 2 */
	(int (*)())	NULL,		/* 3 */
	(int (*)())	NULL,		/* 3 */
	(int (*)())	NULL,		/* 4 */
	(int (*)())	NULL,		/* 4 */
	(int (*)())	NULL,		/* 5 */
	(int (*)())	NULL,		/* 5 */
	(int (*)())	NULL,		/* 6 */
	(int (*)())	NULL,		/* 6 */
	(int (*)())	NULL,		/* 7 */
	(int (*)())	NULL,		/* 7 */
	(int (*)())	NULL,		/* 8 */
	(int (*)())	NULL,		/* 8 */
	(int (*)())	NULL,		/* 9 */
	(int (*)())	NULL,		/* 9 */
	(int (*)())	NULL,		/* 10 */
	(int (*)())	NULL,		/* 10 */
	(int (*)())	NULL,		/* 11 */
	(int (*)())	NULL,		/* 11 */
	(int (*)())	NULL,		/* 12 */
	(int (*)())	NULL,		/* 12 */
	(int (*)())	NULL,		/* 13 */
	(int (*)())	NULL,		/* 13 */
	(int (*)())	NULL,		/* 14 */
	(int (*)())	NULL,		/* 14 */
	(int (*)())	NULL,		/* 15 */
	(int (*)())	NULL,		/* 15 */
	(int (*)())	NULL,		/* 16 */
	(int (*)())	NULL,		/* 16 */
	(int (*)())	NULL,		/* 17 */
	(int (*)())	NULL,		/* 17 */
	(int (*)())	NULL,		/* 18 */
	(int (*)())	NULL,		/* 18 */
	(int (*)())	NULL,		/* 19 */
	(int (*)())	NULL,		/* 19 */
	(int (*)())	NULL,		/* 20 */
	(int (*)())	NULL,		/* 20 */
	(int (*)())	NULL,		/* 21 */
	(int (*)())	NULL,		/* 21 */
	(int (*)())	NULL,		/* 22 */
	(int (*)())	NULL,		/* 22 */
	(int (*)())	NULL,		/* 23 */
	(int (*)())	NULL,		/* 23 */
	(int (*)())	NULL,		/* 24 */
	(int (*)())	NULL,		/* 24 */
	(int (*)())	NULL,		/* 25 */
	(int (*)())	NULL,		/* 25 */
	(int (*)())	NULL,		/* 26 */
	(int (*)())	NULL,		/* 26 */
	(int (*)())	NULL,		/* 27 */
	(int (*)())	NULL,		/* 27 */
	(int (*)())	NULL,		/* 28 */
	(int (*)())	NULL,		/* 28 */
	(int (*)())	NULL,		/* 29 */
	(int (*)())	NULL,		/* 29 */
	(int (*)())	NULL,		/* 30 */
	(int (*)())	NULL,		/* 30 */
	(int (*)())	NULL,		/* 31 */
	(int (*)())	NULL,		/* 31 */
	(int (*)())	NULL,		/* 32 */
	(int (*)())	NULL,		/* 32 */
	(int (*)())	NULL,		/* 33 */
	(int (*)())	NULL,		/* 33 */
	wsget,		/* 34 */	/* 5000/200 */
	wsput,		/* 34 */
	(int (*)())	NULL,		/* 35 */
	(int (*)())	NULL,		/* 35 */
	(int (*)())	NULL,		/* 36 */
	(int (*)())	NULL,		/* 36 */
	(int (*)())	NULL,		/* 37 */
	(int (*)())	NULL,		/* 37 */
	(int (*)())	NULL,		/* 38 */
	(int (*)())	NULL,		/* 38 */
	(int (*)())	NULL,		/* 39 */
	(int (*)())	NULL,		/* 39 */
	(int (*)())	NULL,		/* 40 */
	(int (*)())	NULL,		/* 40 */
	(int (*)())	NULL,		/* 41 */
	(int (*)())	NULL,		/* 41 */
	(int (*)())	NULL,		/* 42 */
	(int (*)())	NULL,		/* 42 */
	(int (*)())	NULL,		/* 43 */
	(int (*)())	NULL,		/* 43 */
	(int (*)())	NULL,		/* 44 */
	(int (*)())	NULL,		/* 44 */
	wsget,			/* 45 */	/* 5000/240 */
	wsput,			/* 45 */
	(int (*)())	NULL,		/* 46 */
	(int (*)())	NULL,		/* 46 */
};

#define	XSIZE	120
#define	YSIZE	100
#define	CRUFT	24	/* for receiver side mbuf m_pullup? */
#define	LNSPACE	4	/* # bytes for scan line # */
#define	DATASZ	4000
#define	PKTSZ	(CRUFT+LNSPACE+DATASZ)
#define	NBUFS	50	/* # frames */
#define	NPKTS	(((XSIZE*YSIZE)/DATASZ))

int
wsget(sp, data, len)
	struct splice_descriptor *sp;
	void *data;
	int len;
{
	extern int udpcksum;
	extern int donothing();
	extern caddr_t output();

	register i;
	int err;
	struct mbuf *m, *p;

#ifdef NOUDPCKSUM
	udpcksum = 0;	/* turn off udp checksums */
#endif

	/* M_CLUSTERSZ is 4096 bytes on ultrix/mips */

	/*
	 * we know that ultrix udp_output will always allocate
	 * a new mbuf for the header, so we just send data here.
	 */

	for (i = 0; i < NPKTS*NBUFS; i++) {
		/*
		 * get an mbuf to point at our data buffer
		 */

		MGET(m, M_DONTWAIT, MT_DATA);
		if (m == (struct mbuf *) NULL) {
			cprintf("wsget: out of mbufs!\n");
			return;
		}
		p = (struct mbuf *) output(i);

		/*
		 * make it point to our data buffer
		 */

		MCLPUT(m, p); /* CLTYPE2 appears to be used for nfs, but
		 		* it's useful for us here also */

		m->m_cltype = M_CLTYPE2;	/* has cl fn! */

		/* clfun will be called during M_FREE, but here
		 * we're using static data, so can't free that
		 */
		m->m_clfun = donothing;
		m->m_clarg = i;	/* for shits and giggles only */

		m->m_len = PKTSZ;	/* cruft and all */

		/* call udp_usrreq */
		err = (*sp->sp_destsock->so_proto->pr_usrreq)
			(sp->sp_destsock, PRU_SEND, m, 0, 0);

		if (err != 0) {
			cprintf("wsget err=%d\n", err);
		}
		
	}
	cprintf("wsget went %d times\n", i);
	cprintf("wsget sent %d bytes\n", i*PKTSZ);
	return (0);
}

int
donothing()
{
	/* cprintf("nothing..\n"); */
}

extern ws_screens screens[];

int
wsput(sp, data, len)
	struct splice_descriptor *sp;
	void *data;
	int len;
{
	extern int udpcksum;
	int screen = 0;
	register ws_screens *wsp = &screens[screen];
	register ws_depth_descriptor *wdd = wsp->dp;
	register int *startline = (int *) data;

#ifdef NOUDPCKSUM
	udpcksum = 0;
#endif
	if ((((u_long) startline) & 0x3) != 0) {
		uprintf("wsput: misaligned startline: 0x%x\n", startline);
		return;
	}
	*startline = ntohl(*startline);

fstrace_vid(*startline);

	data = (u_char *) data + sizeof(u_long); /* uglix compiler :( */
	if ((*startline < 0) || (*startline > 100)) {
		uprintf("wsput: bad startline %d\n", startline);
		return;
	}
	/*
	 * we have decoded the "packet" by pulling out the
	 * first longword, which will tell us where this
	 * packet of stuff belongs in the window
	 * drawit() takes care of actually displaying the packet
	 * in the window [note the use of physaddr, which
	 * implies we're doing writes at the address
	 * directly where the fb is mapped in the kernel
	 * (at this point in time, I expect the memory map
	 * is direct, so the kern virt. addr for the fb
	 * is the same as it's actual physaddr... need to
	 * check this at some point (kfall, 10/1/92)
	 */

	drawit(data, wdd->physaddr, *startline, len - sizeof(u_long));
}


/*
 * drawit:
 *	code assumes the following constants are accurate
 */


#define	CX_FB_WIDTH	1024	/* # pixels across in CX fb */
#define	CX_FB_HEIGHT	864	/* # pixels up/down in CX fb */
#define	WINDOW_WIDTH	120	/* test window width (bytes) */
#define	WINDOW_HEIGHT	100	/* test window height */

static int
drawit(indata, fbaddr, startline, len)
	void	*indata;	/* words to draw */
	caddr_t	fbaddr;		/* address of frame buffer */
	int	startline;	/* line to begin drawing */
	int	len;		/* number of bytes to draw */
{
	register nlines;
	register char (*fbp)[CX_FB_WIDTH] = (char (*)[CX_FB_WIDTH])
		fbaddr;

	register struct fb_template {
		u_char	dummy[CX_FB_WIDTH];
	} *frame_p = (struct fb_template *) (fbp + startline);

	struct window_template {
		u_char	dummy[WINDOW_WIDTH];
	};

	register struct window_template *inp =
		(struct window_template *) indata;

	/*
	 * loop through the number of horiz. small lines
	 * we have to draw
	 */

	for (nlines = 0; nlines < (len / WINDOW_WIDTH); nlines++) {
		/*
		 * the next struct assignment line moves one
		 * horiz. small window's worth of data from the
		 * input into the frame buffer
		 */

		* (struct window_template *) frame_p = *inp++;

		/*
		 * now, move frame_p to point to the next frame
		 * buffer scan line address
		 */
		
		++frame_p;
	}
}

static caddr_t
output(pktnum)
	int pktnum;
{
	static int calledyet = 0;
	static char buf[NBUFS][NPKTS][PKTSZ];
	register i, j;
	u_long *lp;

	if (!calledyet) {
		register caddr_t	dataptr;
		for (i = 0; i < NBUFS; i++) {
			for (j = 0; j < NPKTS; j++) {
				/* skip cruft and scanline */
				dataptr = (caddr_t)
					&buf[i][j][CRUFT+LNSPACE];
				/* set data parts to i */
				memset(dataptr, (u_char)(i & 0xff),
					DATASZ);

				/* set scanline */
				dataptr = &buf[i][j][CRUFT];
				lp = (u_long *)dataptr;
				*lp = htonl(((j*PKTSZ)/XSIZE));
			}
		}
		calledyet = 1;
	}

	/*
	 * we have the following:
	 *	100 lines, each 100 wide, or 100 bytes/line
	 *	24 extra "overhead" bytes to pad at the beginning
	 *	1 word of scan line info in data region (1st word)
	 *
	 *	break it up as follows:
	 *	for each buffer:
	 *		1st packet: 24 bytes cruft
	 *			     4 bytes line#
	 *			  4000 bytes picture
	 *			  4028 byte write
	 *		2nd packet: 24 bytes cruft
	 *			     4 bytes line#
	 *			  4000 bytes picture
	 *			  4028 byte write
	 *		3rd packet: 24 bytes cruft
	 *			     4 bytes line#
	 *			  4000 bytes picture	(2k used)
	 */

	 {
		u_char	(*p)[PKTSZ];
		p = (u_char (*)[PKTSZ]) buf;
		return ((caddr_t) (p + pktnum));
	}
}

memset(sp, c, n)
register char *sp, c;
register int n;
{
	while (--n >= 0)
		*sp++ = c;
	return;
}
