
/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * splice_bread.c
 *	modified versions of bread and breada which don't block
 *	awaiting I/O completion
 *
 *	also, they call setup_callback() to make sure the
 *	splice handler is invoked when the bio is complete
 */

#include "../machine/pte.h"
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/proc.h"
#include "../h/seg.h"
#include "../h/smp_lock.h"
#include "../h/vm.h"
#include "../h/trace.h"
#include "../h/gnode.h"
#include "../h/mount.h"
#include "../h/cpudata.h"
#include "../h/mbuf.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/errno.h"

#include "../net/net/route.h"
#include "../net/net/if.h"

#include "../net/netinet/in.h"
#include "../net/netinet/in_pcb.h"
#include "../net/netinet/in_systm.h"
#include "../net/netinet/in_var.h"
#include "../net/netinet/ip.h"
#include "../net/netinet/ip_var.h"

#include "../csl/splice.h"

#undef SPLICE_DEBUG
extern struct bufstats bufstats;
void setup_callback();
extern int splice_read_handler();

/*
 * Read in (if necessary) the block and return a buffer pointer.
 * modified to not call biowait() and to call setup_callback()
 */

struct buf *
splice_bread(dev, lblkno, blkno, size, gp, sp, handler)
	dev_t dev;
	register daddr_t lblkno, blkno;
	register int size;
	register struct gnode *gp;
	register struct splice_descriptor *sp;
	int (*handler)();
{
	register struct buf *bp;
	int saveaffinity;

	if (size == 0)
		panic("splice_bread: size 0");

loop:
	bp = splice_getblk(dev, blkno, size, gp);

	if (bp == NULL)
		return(NULL);

	bp->b_lblkno = lblkno;
	if (bp->b_flags&B_DONE) {

#ifdef	SPLICE_DEBUG
cprintf("splice_bread: cache read hit, buf@0x%x\n", bp);
#endif
		++bufstats.readhit;
		return(bp);
	}

	if (bp->b_flags&B_DELWRI) {

#ifdef	SPLICE_DEBUG
cprintf("splice_bread: delayed-write: buf@0x%x\n", bp);
#endif
panic("splice_bread: delayed write buf");
		++bufstats.delwrite;
		bwrite(bp);
		goto loop;
	}

	/*
	 * POLICY::: if there's a handler, install it
	 * otherwise just set the ASYNC bit, and busywait
	 */

	if (handler != (int (*)()) NULL)
		setup_callback(bp, sp, handler);
	else
		bp->b_flags |= B_ASYNC;

	bp->b_flags |= B_READ;
	if (bp->b_bcount > bp->b_bufsize)
		panic("splice_bread");

	STRATEGY(gp, dev, bp, saveaffinity);

	++bufstats.readmiss;
	u.u_ru.ru_inblock++;		/* pay for read */
	/* biowait() removed kfall */
	return (NULL);
}


/*
 * Read in the block, like bread, but also start I/O on the
 * read-ahead block (which is not allocated to the caller)
 * modified to not call biowait()
 */
struct buf *
splice_breada(dev, blkno, size, rablkno, rabsize, gp, sp)
	dev_t dev;
	register daddr_t blkno; 
	register int size;
	register daddr_t rablkno;
	register struct gnode *gp;
	register struct splice_descriptor *sp;
	int rabsize;
{
	register struct buf *bp, *rabp;
	int saveaffinity;

	bp = NULL;

#ifdef SPLICE_DEBUG
cprintf("splice_breada:\n");
#endif

	if (!incore(dev, blkno, gp)) {
		bp = getblk(dev, blkno, size, gp);

		if ((bp->b_flags&B_DONE) == 0) {
			setup_callback(bp, sp, splice_read_handler);
			bp->b_flags |= B_READ;
			if (bp->b_bcount > bp->b_bufsize)
				panic("splice_breada");
			STRATEGY(gp, dev, bp, saveaffinity);
			++bufstats.readmiss;
			u.u_ru.ru_inblock++;		/* pay for read */
		} else {
			trace(TR_BREADHIT, dev, blkno);
			++bufstats.readhit;
		}
	}

	/*
	 * If there's a read-ahead block, start i/o
	 * on it also (as above).
	 */
	if (rablkno) {
		if (!incore(dev, rablkno, gp)) {
			rabp = getblk(dev, rablkno, rabsize, gp);
			if (rabp->b_flags & B_DONE) {
				brelse(rabp);
				++bufstats.readahit;
			} else {
				rabp->b_flags |= B_READ|B_ASYNC;
				if (rabp->b_bcount > rabp->b_bufsize)
					panic("splice_breadrabp");
				STRATEGY(gp, dev, rabp, saveaffinity);
				++bufstats.readamiss;
				u.u_ru.ru_inblock++;    /* pay in advance */
			}
		}
		else {
			trace(TR_BREADHITRA, dev, blkno);
			++bufstats.readahit;
		}
	}

	/*
	 * If block was in core, let bread get it.
	 * If block wasn't in core, then the read was started
	 * above, and just wait for it.
	 */
	if (bp == NULL)
		return (splice_bread(dev, blkno, size, gp, sp));
	/* removed biowait() */
	return (bp);
}
