
/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * splice_read:
 *	basically emulates the read part of rwip...(ufs_rwgp in ultrix)
 *	given an offset into a file (must be block aligned),
 *	read in the given block into the buffer cache
 *	and return; will fill in the buffer pointer, and return
 *	a value indicating if there was a cache hit or miss
 *	on a cache miss, the read_handler will be invoked after
 *	the read strategy routine  completes the read io
 *
 * Return values are found defined in splice.h
 */

#define	SPLICE_DEBUG	1


#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/sysproc.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/smp_lock.h"
#include "../h/buf.h"
#include "../h/mbuf.h"
#include "../h/un.h"
#include "../h/conf.h"
#include "../h/mount.h"
#include "../h/gnode_common.h"
#include "../h/gnode.h"
#include "../h/buf.h"
#include "../h/fs.h"
#include "../fs/ufs/fs.h"
#include "../fs/ufs/ufs_inode.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/stat.h"
#include "../h/ioctl.h"
#include "../h/uio.h"
#include "../net/net/route.h"
#include "../net/netinet/in.h"
#include "../net/net/if.h"
#include "../net/netinet/in_pcb.h"
#include "../h/kmalloc.h"
#include "splice.h"

/*
 * see splice.h for definitions of return values
 */

int
splice_read(sp, rbp)
	register struct splice_descriptor *sp;	/* splice desc */
	struct buf **rbp;		/* returned buf ptr */
{
	register struct gnode *gp = sp->sp_srcgp;	/* src gp */

	dev_t srcdev = (dev_t) gp->g_dev;	/* src device */
	struct fs *fs = FS(gp);			/* src filesys */
	long bsize = fs->fs_bsize;		/* src fs lblk size */
	int offset;
	daddr_t lbn;
	daddr_t	pbn;				/* src phys blkno */

	int size, n, s;
	struct buf *bp;

	*rbp = NULL;

	s = splbio();
	n = MIN((unsigned) bsize, (gp->g_size - offset) );
	offset = sp->sp_srcoffset;
	sp->sp_srcoffset += n;

	/*
	 * If we're at the end, return EOF
	 */

	if (offset >= gp->g_size) {
		sp->sp_srcoffset -= n;
		splx(s);
		return (SPLICE_READ_EOF);
	}
	splx(s);

	/*
	 * we need either an entire fs block, or just the
	 * last part if in the last block
	 */

	lbn = lblkno(fs, offset);
	size = blksize(fs, gp, lbn);	/* size of *this* file blk */ 
					/* could be frag or fs block */
					/* used for calls to getblk */
					/* see ufs/fs.h */


	pbn = sp->sp_srcblockmap[(int) lbn];
	if ((int)pbn <= 0)
		panic("splice_read: pbn out of range");

	s = splbio();
	sp->sp_oreads++;
	splx(s);
	/* splice_bread will return NULL on cache miss */
	bp = splice_bread(srcdev, lbn, pbn, size,
		(struct gnode *) NULL, sp, splice_read_handler);

	if (bp) {
		s = splbio();
		sp->sp_oreads--;
		splx(s);
	}

#ifdef notdef
	if (gp->g_lastr + 1 == lbn)
		bp = splice_breada(
			srcdev, pbn, size, rablock, rasize,
			(struct gnode *) NULL, sp);
	else
		bp = splice_bread(
			srcdev, pbn, size,
			(struct gnode *) NULL, sp);
#endif

	if (bp && (bp->b_flags & B_ERROR)) {
		brelse(bp);
		return (SPLICE_READ_ERROR);
	}

	/*
	 * bread(a) will have returned bp==NULL if the
	 * block was not cached.  In such a case, it will
	 * have initiated a read via the strategy routines
	 */

#ifdef	SPLICE_DEBUG
cprintf("splice_read: finished bread: returning[bp@0x%x]\n", bp);
#endif

	gp->g_lastr = lbn;

	/*
	 * on cache hit, fill in the fields
	 */

	if (bp) {
		*rbp = bp;
		bp->b_splicep = (u_char *) sp;
		return (SPLICE_READ_HIT);
	}

	/*
	 * if we get here, the strategy routine has
	 * been called to fetch the block in question
	 */

	return (SPLICE_READ_MISS);
}


/*
 * splice_read_handler()
 *	gets invoked via the buffer callback [biodone()]
 *	we will have completed a bread at this point,
 *	and must schedule the bwrite.
 *	After doing the write, call splice_rw to start the procedure
 *	all over again.
 */

#include "../h/ksched.h"

int
splice_read_handler(bp)
	struct buf *bp;
{
	int rval, i;
	struct splice_descriptor *sp = (struct splice_descriptor *)
			bp->b_splicep;

	struct kschedblk	kblk; /* ok to be an automatic */

	if (--sp->sp_oreads < 0)
		panic("splice_read_handler: negative oreads"); 


#ifdef notdef
	/* if we are too low, give us some more */
	if (sp->sp_oreads < SPLICE_MINREADS) {
		for (i = SPLICE_MAXOREADS; i; i--)
			if (splice_rw(sp) == 1)
				break;
	}
#endif

#ifdef notdef
	if (bp->b_lblkno == 0) {
		for (i = 0; i < SPLICE_MAXOREADS; i++) {
			/* if EOF on the read, break out early */
			if (splice_rw(sp) == 1)
				break;
		}
	}
#endif

#ifdef SPLICE_DEBUG
cprintf("splice_read_handler: buf@0x%x, flags=0x%x, nreads=%d, nwrites=%d\n",
bp, bp->b_flags, sp->sp_oreads, sp->sp_owrites);
#endif

	/*
	 * this code looks reversed, but note that in the
	 * ksched routine entries with the same timeout are put in
	 * LIFO, so we need to swap the order here
	 */

#ifdef notdef
	if ((sp->sp_owrites < SPLICE_MAXOWRITES)) {
		/* schedule subsequent reads */
		kblk.arg = (caddr_t) sp;
		kblk.func = splice_rw;
		ksched(&kblk);
	}
#endif

	/* schedule a call to the writer */
	kblk.arg = (caddr_t) bp;
	kblk.func = (void (*)()) splice_write;
	sp->sp_owrites++;
	ksched(&kblk);

	return;
}

/*
 * network upcall--
 *	called by udp_input in udp_usrreq.c 
 *	we are limited to inet domain due to sockaddr_in
 *	the mbuf chain *includes* the struct udpiphdr
 *
 * regarding udp checksums:
 *	udp checksums occur when "udpcksum" is set to 1, which
 *	it is by default.
 *
 *	in udp_input (udp_usrreq.c, line 241) "udpcksum" is checked
 *	as to determine whether to perform the cksum on input
 *
 *	note that splice_net_upcall is called *after* this check
 *	so a correct udp cksum is ensured, provided "udpcksum" is one
 *
 *	on the outgoing trip, in udp_output (udp_usrreq, line 487)
 *	"udpcksum" is once again checked to determine whether to
 *	compute the checksum.  Note that this vers of UNIX
 *	always computes the udp checksum provided udpcksum is set.
 *	(even tho the udp spec does support a "no checksum" option
 */

splice_net_upcall(inp, srcaddr, m)
	register struct inpcb *inp;
	register struct sockaddr_in *srcaddr;
	register struct mbuf *m;
{
	struct splice_descriptor *sp;
	int error; 


	if ((sp = (struct splice_descriptor *)inp->inp_splicep) == NULL)
		panic("udp splice with NULL splice ptr");

	if (ISSOCK(sp->sp_destfile)) {
ntrace_rec("WSP", m);
		/*
		 * for now, make the 2nd to last arg be zero,
		 * indicating that there should already be
		 * a far-side address.  That is, the socket
		 * should already have been connected
		 */
		error = 0;
		error = (*sp->sp_destsock->so_proto->pr_usrreq)
			(sp->sp_destsock, PRU_SEND, m, 0, 0);
		if (error) {
#ifdef	SPLICE_DEBUG
cprintf("splice_upc: ERR code %d", error);
/*
if (m) {
m->m_next = NULL;
walk_mbuf_chain(m);
}
*/
#endif
			m_freem(m);	/* free whole chain */
		}
		return;
	} else if (ISINODE(sp->sp_destfile)) {
		/* check if it's a char. special device */
		if (ISCHR(sp->sp_destgp) && sp->sp_cdevput) {
			/*
			 * XXX: terrible code here!
			 * for some reason, 1st mbuf of chain
			 * will have 24 data bytes in in,
			 * then 2nd will have the rest (why??)
			 * mysterious m_pullup?
			 */

			struct mbuf *mnext = m->m_next;
			if (mnext) {
				(*sp->sp_cdevput)(sp,
					mtod(mnext, caddr_t),
					mnext->m_len);
			} else {
/* walk_mbuf_chain(m); */
			(*sp->sp_cdevput)(sp, mtod(m, caddr_t),
				(int) m->m_len);
			}
		} else if (ISREG(sp->sp_destgp)) {
			/* not implemented */
		}
		m_freem(m);
	}
}

walk_mbuf_chain(m)
	struct mbuf *m;
{
	while (m) {
		cprintf("\nm:0x%x, m_next=0x%x, m_len=%d, m_type=%d",
			m, m->m_next, m->m_len, m->m_type);
		m = m->m_next;
	}
}
	
